! $Id: omegalib.nw,v 1.114 2004/04/04 22:34:57 ohl Exp $
!
! Copyright (C) 2000-2002 by Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
! and others
!
! O'Mega is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! O'Mega is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
program test_omega95_bispinors
  use omega_kinds
  use omega95_bispinors
  use omega_vspinor_polarizations
  use omega_testtools
  implicit none
  integer :: i, j
  real(kind=omega_prec) :: m, pabs, qabs, w
  real(kind=omega_prec), dimension(4) :: r
  complex(kind=omega_prec) :: one
  type(momentum) :: p, q
  type(vector) :: vp, vq
  type(vectorspinor) :: testv, dum_u, dum_v
  call random_seed ()
  one = 1
  w = 0
  m = 13
  pabs = 42
  qabs = 137
  call random_momentum (p, pabs, m)
  call random_momentum (q, qabs, m)
  call random_number (r)
  do i = 1, 4
     testv%psi(1)%a(i) = (0, 0)
  end do
  do i = 2, 3
     do j = 1, 4
        testv%psi(i)%a(j) = cmplx (10.0_omega_prec * r(j))
    end do
  end do
  testv%psi(4)%a(1) = 1
  testv%psi(4)%a(1) = (0, 2.0_omega_prec)
  testv%psi(4)%a(1) = 1
  testv%psi(4)%a(1) = (3.0_omega_prec, 0)
  vp = p
  vq = q
  dum_u = ueps(m,p,2)
  dum_v = veps(m,p,2)
print *, "*** Checking the equations of motion ***:"
call expect (abs(f_vf(one,vp,u(m,p,+1))-m*u(m,p,+1)), 0, "|[p-m]u(+)|=0")
call expect (abs(f_vf(one,vp,u(m,p,-1))-m*u(m,p,-1)), 0, "|[p-m]u(-)|=0")
call expect (abs(f_vf(one,vp,v(m,p,+1))+m*v(m,p,+1)), 0, "|[p+m]v(+)|=0")
call expect (abs(f_vf(one,vp,v(m,p,-1))+m*v(m,p,-1)), 0, "|[p+m]v(-)|=0")
print *, "*** Checking the normalization ***:"
call expect (s_ff(one,v(m,p,+1),u(m,p,+1)), +2*m, "ubar(+)*u(+)=+2m")
call expect (s_ff(one,v(m,p,-1),u(m,p,-1)), +2*m, "ubar(-)*u(-)=+2m")
call expect (s_ff(one,u(m,p,+1),v(m,p,+1)), -2*m, "vbar(+)*v(+)=-2m")
call expect (s_ff(one,u(m,p,-1),v(m,p,-1)), -2*m, "vbar(-)*v(-)=-2m")
call expect (s_ff(one,v(m,p,+1),v(m,p,+1)),    0, "ubar(+)*v(+)=0  ")
call expect (s_ff(one,v(m,p,-1),v(m,p,-1)),    0, "ubar(-)*v(-)=0  ")
call expect (s_ff(one,u(m,p,+1),u(m,p,+1)),    0, "vbar(+)*u(+)=0  ")
call expect (s_ff(one,u(m,p,-1),u(m,p,-1)),    0, "vbar(-)*u(-)=0  ")
print *, "*** Checking the currents ***:"
call expect (abs(v_ff(one,v(m,p,+1),u(m,p,+1))-2*vp), 0, "ubar(+).V.u(+)=2p")
call expect (abs(v_ff(one,v(m,p,-1),u(m,p,-1))-2*vp), 0, "ubar(-).V.u(-)=2p")
call expect (abs(v_ff(one,u(m,p,+1),v(m,p,+1))-2*vp), 0, "vbar(+).V.v(+)=2p")
call expect (abs(v_ff(one,u(m,p,-1),v(m,p,-1))-2*vp), 0, "vbar(-).V.v(-)=2p")
print *, "*** Checking current conservation ***:"
call expect ((vp-vq)*v_ff(one,v(m,p,+1),u(m,q,+1)), 0, "d(ubar(+).V.u(+))=0")
call expect ((vp-vq)*v_ff(one,v(m,p,-1),u(m,q,-1)), 0, "d(ubar(-).V.u(-))=0")
call expect ((vp-vq)*v_ff(one,u(m,p,+1),v(m,q,+1)), 0, "d(vbar(+).V.v(+))=0")
call expect ((vp-vq)*v_ff(one,u(m,p,-1),v(m,q,-1)), 0, "d(vbar(-).V.v(-))=0")
if (m == 0) then
   print *, "*** Checking axial current conservation ***:"
   call expect ((vp-vq)*a_ff(one,v(m,p,+1),u(m,q,+1)), 0, "d(ubar(+).A.u(+))=0")
   call expect ((vp-vq)*a_ff(one,v(m,p,-1),u(m,q,-1)), 0, "d(ubar(-).A.u(-))=0")
   call expect ((vp-vq)*a_ff(one,u(m,p,+1),v(m,q,+1)), 0, "d(vbar(+).A.v(+))=0")
   call expect ((vp-vq)*a_ff(one,u(m,p,-1),v(m,q,-1)), 0, "d(vbar(-).A.v(-))=0")
end if
print *, "*** Checking polarization vectors: ***"
call expect (conjg(eps(m,p, 1))*eps(m,p, 1), -1, "e( 1).e( 1)=-1")
call expect (conjg(eps(m,p, 1))*eps(m,p,-1),  0, "e( 1).e(-1)= 0")
call expect (conjg(eps(m,p,-1))*eps(m,p, 1),  0, "e(-1).e( 1)= 0")
call expect (conjg(eps(m,p,-1))*eps(m,p,-1), -1, "e(-1).e(-1)=-1")
call expect (                 p*eps(m,p, 1),  0, "    p.e( 1)= 0")
call expect (                 p*eps(m,p,-1),  0, "    p.e(-1)= 0")
if (m > 0) then
   call expect (conjg(eps(m,p, 1))*eps(m,p, 0),  0, "e( 1).e( 0)= 0")
   call expect (conjg(eps(m,p, 0))*eps(m,p, 1),  0, "e( 0).e( 1)= 0")
   call expect (conjg(eps(m,p, 0))*eps(m,p, 0), -1, "e( 0).e( 0)=-1")
   call expect (conjg(eps(m,p, 0))*eps(m,p,-1),  0, "e( 0).e(-1)= 0")
   call expect (conjg(eps(m,p,-1))*eps(m,p, 0),  0, "e(-1).e( 0)= 0")
   call expect (                 p*eps(m,p, 0),  0, "    p.e( 0)= 0")
end if
print *, "*** Checking polarization vectorspinors: ***"
call expect (abs(p * ueps(m, p,  2)),  0, "p.ueps ( 2)= 0")
call expect (abs(p * ueps(m, p,  1)),  0, "p.ueps ( 1)= 0")
call expect (abs(p * ueps(m, p, -1)),  0, "p.ueps (-1)= 0")
call expect (abs(p * ueps(m, p, -2)),  0, "p.ueps (-2)= 0")
call expect (abs(p * veps(m, p,  2)),  0, "p.veps ( 2)= 0")
call expect (abs(p * veps(m, p,  1)),  0, "p.veps ( 1)= 0")
call expect (abs(p * veps(m, p, -1)),  0, "p.veps (-1)= 0")
call expect (abs(p * veps(m, p, -2)),  0, "p.veps (-2)= 0")
print *, "*** Checking the irreducibility condition: ***"
call expect (abs(f_potgr (one, one, ueps(m, p,  2))),  0, "g.ueps ( 2)")
call expect (abs(f_potgr (one, one, ueps(m, p,  1))),  0, "g.ueps ( 1)")
call expect (abs(f_potgr (one, one, ueps(m, p, -1))),  0, "g.ueps (-1)")
call expect (abs(f_potgr (one, one, ueps(m, p, -2))),  0, "g.ueps (-2)")
call expect (abs(f_potgr (one, one, veps(m, p,  2))),  0, "g.veps ( 2)")
call expect (abs(f_potgr (one, one, veps(m, p,  1))),  0, "g.veps ( 1)")
call expect (abs(f_potgr (one, one, veps(m, p, -1))),  0, "g.veps (-1)")
call expect (abs(f_potgr (one, one, veps(m, p, -2))),  0, "g.veps (-2)")
print *, "*** Testing the gravitino propagator: ***"
print *, "Transversality:"
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,testv))), 0, "p.pr.test")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,ueps(m,p,2)))),  0, "p.pr.ueps ( 2)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,ueps(m,p,1)))),  0, "p.pr.ueps ( 1)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,ueps(m,p,-1)))), 0, "p.pr.ueps (-1)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,ueps(m,p,-2)))), 0, "p.pr.ueps (-2)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,veps(m,p,2)))),  0, "p.pr.veps ( 2)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,veps(m,p,1)))),  0, "p.pr.veps ( 1)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,veps(m,p,-1)))), 0, "p.pr.veps (-1)")
call expect (abs(p * (cmplx (p*p - m**2, m*w, kind=omega_prec) * &
             pr_grav(p,m,w,veps(m,p,-2)))), 0, "p.pr.veps (-2)")
print *, "Irreducibility:"
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,testv)))), 0, "g.pr.test")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,ueps(m,p,2))))), 0, &
             "g.pr.ueps ( 2)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,ueps(m,p,1))))), 0, &
             "g.pr.ueps ( 1)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,ueps(m,p,-1))))), 0, &
             "g.pr.ueps (-1)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,ueps(m,p,-2))))), 0, &
             "g.pr.ueps (-2)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,veps(m,p,2))))), 0, &
             "g.pr.veps ( 2)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,veps(m,p,1))))), 0, &
             "g.pr.veps ( 1)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,veps(m,p,-1))))), 0, &
             "g.pr.veps (-1)")
call expect (abs(f_potgr (one, one, (cmplx (p*p - m**2, m*w, & 
             kind=omega_prec) * pr_grav(p,m,w,veps(m,p,-2))))), 0, &
             "g.pr.veps (-2)")
end program test_omega95_bispinors
